# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

from django.db import models
from django.utils.translation import gettext_lazy as _

from quartet_epcis.models import abstractmodels


class Entry(abstractmodels.UUIDModel):
    '''
    Represents an entry in the general ledger of serialized items and/or
    logical numbers used for serialized goods processing.
    '''
    identifier = models.CharField(
        max_length=150,
        null=False,
        help_text=_('The primary unique id for the entry.'),
        verbose_name=_('EPC URN'),
        db_index=True,
        unique=True
    )

    def __str__(self):
        return self.identifier

    class Meta:
        verbose_name = _('Entry')
        verbose_name_plural = _('Entries')
        app_label = 'quartet_epcis'


class EntryEvent(models.Model):
    '''
    An intersection entity for events and entries.
    '''
    event = models.ForeignKey(
        'quartet_epcis.Event',
        null=False,
        help_text=_('The UUID of the event.'),
        verbose_name=_('Event ID'),
        db_index=True,
        on_delete=models.CASCADE
    )
    entry = models.ForeignKey(
        'quartet_epcis.Entry',
        null=False,
        help_text=_('The Unique ID of the Entry'),
        verbose_name=_('Entry ID'),
        db_index=True,
        on_delete=models.CASCADE
    )
    identifier = models.CharField(
        max_length=150,
        null=False,
        help_text=_('A redundant entry ID entry for fast event composition.'),
        verbose_name=_('EPC URN'),
    )
    is_parent = models.BooleanField(
        default=False,
        help_text=_('Whether or not this entry was the parent of it\'s '
                    'constituent event.'),
        verbose_name=_('Is Event Parent')
    )
    output = models.BooleanField(
        default=False,
        help_text=_('Whether or not the entry was the output of '
                    'a Transformation event.'),
        verbose_name=_('Transformation Output')
    )

    def __str__(self):
        return '{0}:{1}'.format(self.entry_id, event_id)

    class Meta:
        verbose_name = _('Entry Event Record')
        verbose_name_plural = _('Entry Event Records')
        index_together = ["event", "entry"]
        app_label = 'quartet_epcis'



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
import logging
from typing import List
from eparsecis.eparsecis import EPCISParser
from quartet_epcis.models import events, entries, choices, headers
from EPCPyYes.core.v1_2 import events as yes_events
from EPCPyYes.core.SBDH import template_sbdh
from django.db import transaction

logger = logging.getLogger('quartet_epcis')
biz_xact_list = List[yes_events.BusinessTransaction]
ilmd_list = List[yes_events.InstanceLotMasterDataAttribute]
source_list = List[yes_events.Source]
destination_list = List[yes_events.Destination]


class QuartetParser(EPCISParser):
    def __init__(self, stream, event_cache_size: int = 1024):
        '''
        Initializes a new QuartetParser.  Item entries and events will
        be cached in memory until either the event_cache_size or
        entry_cache_size limit is reached- whichever occurs first.

        Be aware that the size of the cached database entries may bump up
        against the size limits of query caches in certain RDBMS systems.

        :param stream: The EPCIS stream to parse.
        :param event_cache_size: defaults to 1024.  The number of events
        to cache in memory before pushing to the back-end datastore.
        '''
        super().__init__(stream)
        self.event_cache = []
        self.entry_cache = {}
        self.quantity_element_cache = []
        self.error_declaration_cache = []
        self.business_transaction_cache = []
        self.ilmd_cache = []
        self.source_cache = []
        self.destination_cache = []
        self.entry_event_cache = []
        self.event_cache_size = event_cache_size
        self.source_event_cache = []
        self.destination_event_cache = []
        self._message = None

    def parse(self):
        '''
        Creates the message for use in associating events and then
        executes the base-class parse function.
        :return: returns the message id created by the parsing of tbe
        inbound data.  See the headers.Message model in the models
        package.
        '''
        self._message = headers.Message()
        self._message.save()
        super().parse()
        self.clear_cache()
        return self._message.id

    def handle_sbdh(self,
                    header: template_sbdh.StandardBusinessDocumentHeader):
        db_header = headers.SBDH()
        db_header.message = self._message
        db_sbdh_id = headers.DocumentIdentification()
        logger.debug('Saving the document identification data.')
        db_sbdh_id.document_type = header.document_identification.document_type
        db_sbdh_id.creation_date_and_time = \
            header.document_identification.creation_date_and_time
        db_sbdh_id.multiple_type = header.document_identification.multiple_type
        db_sbdh_id.instance_identifier = \
            header.document_identification.instance_identifier
        db_sbdh_id.type_version = header.document_identification.type_version
        db_sbdh_id.standard = header.document_identification.standard
        db_sbdh_id.save()
        db_header.document_identification = db_sbdh_id
        logger.debug('Document identification is saved, looking for '
                     'partner data')
        db_header.save()
        partner_cache = []
        if header.partners:
            for partner in header.partners:
                db_partner = headers.Partner(
                    partner_type=partner.partner_type,
                    header=db_header
                )
                if partner.partner_id:
                    db_partner.authority = partner.partner_id.authority
                    db_partner.identifier = partner.partner_id.value
                if partner.has_contact_info:
                    db_partner.contact = partner.contact
                    db_partner.email_address = partner.email_address
                    db_partner.fax_number = partner.fax_number
                    db_partner.telephone_number = partner.telephone_number
                    db_partner.contact_type_identifier = \
                        partner.contact_type_identifier
                    partner_cache.append(db_partner)
                logger.debug('Adding partner to the sbdh model instance.')
        [p.save() for p in partner_cache]


    def handle_transaction_event(
        self,
        epcis_event: yes_events.TransactionEvent
    ):
        '''
        Called whenever the parser has completed parsing a TransactionEvent
        within an EPCIS xml structure.

        :param epcis_event: An EPCPyYes TransactionEvent class instance.
        :return: None
        '''
        logger.debug('Handling a transaction event.')
        db_event = self.get_db_event(epcis_event)
        db_event.type = choices.EventTypeChoicesEnum.TRANSACTION.value
        self.handle_entries(db_event, epcis_event.epc_list)
        if epcis_event.parent_id:
            self.handle_top_level_id(epcis_event.parent_id, db_event)
        self.handle_common_elements(db_event, epcis_event)
        self.event_cache.append(db_event)
        if len(self.event_cache) >= self.event_cache_size:
            self.clear_cache()

    def handle_top_level_id(self, top_id, db_event):
        '''
        For both transaction and aggregation events.  Will store the parent
        and or/top level id as an Entry in the entry cache.
        '''
        # check the cache
        entry = self.entry_cache.get(
            top_id,
            None,
        )
        # not in the cache then create and put in the cache
        if not entry:
            entry = entries.Entry.objects.get_or_create(identifier=top_id)[0]
            self.entry_cache[entry.identifier] = entry

        entryevent = entries.EntryEvent(entry=entry,
                                        event=db_event,
                                        identifier=top_id,
                                        is_parent=True)
        self.entry_event_cache.append(entryevent)
        logger.debug('Cached Entry for top id %s', top_id)

    def handle_aggregation_event(
        self,
        epcis_event: yes_events.AggregationEvent
    ):
        '''
        Executed when an AggregationEvent xml structure has finished parsing.

        :param epcis_event: An EPCPyYes AggregationEvent instance.
        '''
        logger.debug('Handling ann aggregation event.')
        db_event = self.get_db_event(epcis_event)
        db_event.type = choices.EventTypeChoicesEnum.AGGREGATION.value
        self.handle_entries(db_event, epcis_event.child_epcs)
        self.handle_common_elements(db_event, epcis_event)
        self.handle_top_level_id(epcis_event.parent_id, db_event)
        self.event_cache.append(db_event)

    def handle_object_event(self, epcis_event: yes_events.ObjectEvent):
        '''
        Executed when an ObjectEvent xml structure has finished parsing.
        The EParseCIS library will pass in an EPCPyYes ObjectEvent
        class instance for use.

        :param epcis_event: The EPCPyYes ObjectEvent.
        '''
        logger.debug('Handling an ObjectEvent...')
        db_event = self.get_db_event(epcis_event)
        db_event.type = choices.EventTypeChoicesEnum.OBJECT.value
        self.handle_entries(db_event, epcis_event.epc_list)
        self.handle_common_elements(db_event, epcis_event)
        self.handle_ilmd(db_event.id, epcis_event.ilmd)
        self.event_cache.append(db_event)

    def handle_transformation_event(
        self,
        epcis_event: yes_events.TransformationEvent
    ):
        '''
        Executed when a TransformationEvent xml element has completed parsing
        into a valid EPCPyYes TransformationEvent
        :param epcis_event: The EPCPyYes TransformationEvent
        :return: None
        '''
        logger.debug('Handling a TransformationEvent...')
        db_event = self.get_db_event(epcis_event)
        db_event.type = choices.EventTypeChoicesEnum.TRANSFORMATION.value
        self.handle_common_elements(db_event, epcis_event)
        self.handle_entries(db_event, epcis_event.input_epc_list)
        self.handle_entries(db_event, epcis_event.output_epc_list,
                            output=True)
        self.handle_ilmd(db_event.id, epcis_event.ilmd)
        self.event_cache.append(db_event)

    def handle_common_elements(
        self,
        db_event: events.Event,
        epcis_event: yes_events.EPCISBusinessEvent
    ):
        '''
        Helper function to handle the common elements across the Object,
        Transaction and Aggregation event models.
        :param epcis_event: The Event django database model instance.
        :param db_event: The EPCPyYes EPCISBusinessEvent instance being passed
        to the parser.
        '''
        self.handle_business_transactions(
            db_event.id,
            epcis_event.business_transaction_list
        )
        self.handle_source_list(
            db_event.id,
            epcis_event.source_list
        )
        self.handle_destination_list(
            db_event.id,
            epcis_event.destination_list
        )
        if isinstance(epcis_event, yes_events.TransactionEvent) or \
            isinstance(epcis_event, yes_events.ObjectEvent):
            self.handle_quantity_elements(
                db_event.id,
                epcis_event.quantity_list
            )
        elif isinstance(epcis_event, yes_events.AggregationEvent):
            self.handle_quantity_elements(
                db_event.id,
                epcis_event.child_quantity_list
            )
        elif isinstance(epcis_event, yes_events.TransformationEvent):
            self.handle_quantity_elements(
                db_event.id,
                epcis_event.input_quantity_list
            )
            self.handle_quantity_elements(
                db_event.id,
                epcis_event.output_quantity_list,
                is_output=True
            )

    def get_db_event(self, epcis_event):
        '''
        Creates an Event database model with the common fields and returns.
        :param epcis_event: The EPCPyYes event.
        :return: A new Event model instance.
        '''
        db_event = events.Event()
        if not isinstance(epcis_event, yes_events.TransformationEvent):
            db_event.action = epcis_event.action
        db_event.biz_location = epcis_event.biz_location or None
        db_event.biz_step = epcis_event.biz_step or None
        db_event.disposition = epcis_event.disposition or None
        db_event.read_point = epcis_event.read_point or None
        db_event.event_time = epcis_event.event_time
        db_event.event_timezone_offset = epcis_event.event_timezone_offset \
                                         or None
        db_event.event_id = epcis_event.event_id or None
        db_event.record_time = epcis_event.record_time or None
        db_event.message_id = self._message.id
        if epcis_event.error_declaration:
            self.handle_error_declaration(
                db_event.id,
                epcis_event.error_declaration
            )
        return db_event

    def handle_entries(
        self, db_event: str, epc_list: [],
        output: bool = False
    ):
        '''
        Gets the EPCs from the event and caches them for storage in the
        back-end.  Then creates the EntryEvent intersection entity records
        and appends them for storage as well.
        :param db_event: The unique id of the event
        :param epc_list: A list of epcs to be cached.
        :return:
        '''
        logging.debug('Processing epc list %s', epc_list)
        for epc in epc_list:
            entry = self.entry_cache.get(epc)
            if not entry:
                entry = entries.Entry.objects.get_or_create(identifier=epc)[0]
                self.entry_cache[entry.identifier] = entry
            entryevent = entries.EntryEvent(entry=entry,
                                            event=db_event,
                                            identifier=epc,
                                            output=output)
            self.entry_event_cache.append(entryevent)

    def handle_error_declaration(
        self,
        db_event_id: str,
        error_declaration: yes_events.ErrorDeclaration,
    ):
        '''
        Creates an ErrorDeclaration database model and caches
        it for storage in the backend database.
        :param db_event_id: The unique id of the parent event.
        :param error_declaration: The EPCPyYes ErrorDeclaration instance.
        '''
        if isinstance(error_declaration.corrective_event_ids, list):
            logging.debug('Concatenating the corrective event ids')
            cids = ','.join(error_declaration.corrective_event_ids)
        else:
            cids = str(error_declaration.corrective_event_ids)
        ed = events.ErrorDeclaration(
            event_id=db_event_id,
            reason=error_declaration.reason,
            corrective_event_ids=cids,
        )
        self.error_declaration_cache.append(ed)

    def handle_quantity_elements(
        self,
        db_event_id: str,
        quantity_elements: list,
        is_output: bool = False,
    ):
        '''
        Creates a QuantityElement database model and caches it
        for later bulk insert into the backend database.
        :param db_event_id: The source event.
        :param quantity_element: The EPCPyYes QuantityElement
        '''
        for quantity_element in quantity_elements:
            qe = events.QuantityElement(
                event_id=db_event_id,
                epc_class=quantity_element.epc_class,
                quantity=quantity_element.quantity,
                uom=quantity_element.uom,
                is_output=is_output
            )
            logging.debug('Appending quantity element.')
            self.quantity_element_cache.append(qe)

    def handle_business_transactions(self, db_event_id: str,
                                     business_transaction_list: biz_xact_list):
        '''
        Takes each business transaction in the list and creates
        a BusinessTransaction database model instance and caches it
        for bulk insert into the backend database.
        :param db_event_id: The source event primary key/id.
        :param business_transaction_list: A list of EPCPyYes
        BusinessTransactions
        '''
        for transaction in business_transaction_list:
            bt = events.BusinessTransaction(
                event_id=db_event_id,
                biz_transaction=transaction.biz_transaction,
                type=transaction.type
            )
            logging.debug('Appending biz transaction.')
            self.business_transaction_cache.append(bt)

    def handle_ilmd(self, db_event_id: str,
                    ilmd_data: ilmd_list):
        '''
        Takes the ILMD node and creates InstanceLotMasterData model
        instances to be cached for bulk insert into the database.
        :param db_event_id: The source event primary key value.
        :param ilmd_data: The ilmd section of EPCPyYes event.
        '''
        for ilmd in ilmd_data:
            ie = events.InstanceLotMasterData(
                event_id=db_event_id,
                name=str(ilmd.name),
                value=ilmd.value,
            )
            self.ilmd_cache.append(ie)

    def handle_source_list(self, db_event_id: str,
                           sources: source_list):
        '''
        Creates a source database model instance and caches it
        for storage during bulk insert.
        :param db_event_id: The source event primary key.
        :param sources: A list of EPCPyYes Source instances.
        '''
        for source in sources:
            src = events.Source(
                type=source.type,
                source=source.source,
            )
            source_event = events.SourceEvent(source=src,
                                              event_id=db_event_id)
            logger.debug('Adding source %s %s and the source event '
                         'to the cache',
                         source.type, source.source)
            self.source_event_cache.append(source_event)
            self.source_cache.append(src)

    def handle_destination_list(self, db_event_id: str,
                                destinations: destination_list):
        '''
        Creates a destination database model instance and caches
        it for storage during bulk insert.
        :param db_event_id: The source event's primary key.
        :param destinations: A list of EPCPyYes Destination instances.
        '''
        for destination in destinations:
            dest = events.Destination(
                type=destination.type,
                destination=destination.destination
            )
            destination_event = events.DestinationEvent(
                destination=dest,
                event_id=db_event_id
            )
            logger.debug('Adding destination %s %s and the destination event '
                         'to the cache',
                         destination.type, destination.destination)
            self.destination_event_cache.append(destination_event)
            self.destination_cache.append(dest)

    def clear_cache(self):
        '''
        Calls save on all items in the cache
        '''
        with transaction.atomic():
            logger.debug('Clear cache has been called with %s and %i '
                         'in the event and entry caches respectively',
                         len(self.event_cache), len(self.entry_cache))
            events.Event.objects.bulk_create(self.event_cache)
            logger.debug('Clearing out %s number of EntryEvents.',
                         len(self.entry_event_cache))
            entries.EntryEvent.objects.bulk_create(self.entry_event_cache)
            logger.debug('Clearing cache of %s number of quantity elements',
                         len(self.quantity_element_cache))
            events.QuantityElement.objects.bulk_create(
                self.quantity_element_cache
            )
            logger.debug('Clearing cache of %s number of error declarations',
                         len(self.error_declaration_cache))
            events.ErrorDeclaration.objects.bulk_create(
                self.error_declaration_cache
            )
            logger.debug(
                'Clearing the biz transaction cache of %s transactions',
                len(self.business_transaction_cache))
            events.BusinessTransaction.objects.bulk_create(
                self.business_transaction_cache
            )
            logger.debug('Clearing the ILMD cache of %s objects',
                         len(self.ilmd_cache))
            events.InstanceLotMasterData.objects.bulk_create(self.ilmd_cache)
            logger.debug('Clearing out the source cache of %s items',
                         len(self.source_cache))
            events.Source.objects.bulk_create(self.source_cache)
            logger.debug('Clearing out the destination cache of %s items',
                         len(self.destination_cache))
            events.Destination.objects.bulk_create(self.destination_cache)
            logger.debug('Clearing out the source event cache.')
            events.SourceEvent.objects.bulk_create(self.source_event_cache)
            logger.debug('Clearing out the destination event cache.')
            events.DestinationEvent.objects.bulk_create(
                self.destination_event_cache
            )
            logger.debug('Clearing out the cache lists.')
            del self.event_cache[:]
            self.entry_cache.clear()
            del self.entry_event_cache[:]
            del self.error_declaration_cache[:]
            del self.quantity_element_cache[:]
            del self.business_transaction_cache[:]
            del self.ilmd_cache[:]
            del self.source_cache[:]
            del self.destination_cache[:]
            del self.source_event_cache[:]
            del self.destination_event_cache[:]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
import logging
from typing import List
from gettext import gettext as _
from rest_framework import views
from rest_framework.response import Response
from rest_framework.request import Request
from rest_framework.exceptions import NotFound
from rest_framework import status
from EPCPyYes.core.v1_2 import template_events
from quartet_epcis.db_api.queries import EPCISDBProxy
from quartet_epcis.models import events, headers

logger = logging.getLogger(__name__)
EventList = List[events.Event]
proxy = EPCISDBProxy()


class FormatHelperMixin:
    '''
    This mixin helps determine whether or not to call the xml template_event
    functionality or the JSON functionality based on the inbound
    content type of the HTTP request or the explicit format request
    in the HTTP query parameters.
    '''

    def get_formatted_data(self, request: Request, template_event,
                           format=None):
        if 'xml' in request.content_type.lower() or \
            'xml' in request.query_params.get('format', '') or \
            format == 'xml':
            response_data = template_event.render()
        else:
            # else render JSON
            response_data = template_event.render_dict()
        return response_data


class EventDetailView(views.APIView, FormatHelperMixin):
    def get(self, request: Request, format=None, event_id: str = None):
        '''
        Based on an inbound event id, will return the full event data as
        JSON or XML string.  The event_id that is passed in can be either the
        primary key of the event in the database or the actual EPCIS event_id
        value that uniquely identifies the event.

        The URI for a given event is as below (assuming that the urls.py
        was configured for access via `epcis`):

        .. code-block:: text

            # default format is json so format query parameter is optional
            http[s]:/[hostname]:[port]/epcis/event-detail/[event id]/?format=[xml,json]

            # for example
            http://localhost:8000/epcis/event-detail/1d1fa10f-4421-4f2a-af10-4ddb4951807c/?format=json

        :return: A JSON string representing the event.

        '''

        response_data = {}
        try:
            if event_id:
                event = proxy.get_event_by_id(event_id)
                response_data = self.get_formatted_data(request, event, format)
            response = Response(response_data, status=status.HTTP_200_OK)
        except events.Event.DoesNotExist:
            response = Response('Event with id %s could not be '
                                'found' % event_id, status.HTTP_404_NOT_FOUND)
        return response


class EntryEventHistoryView(views.APIView, FormatHelperMixin):
    '''
    Returns all of the events associated with a given EPC/Entry.
    '''

    def get(self, request: Request, format=None, entry_pk: str = None,
            entry_identifier: str = None):
        '''
        Will pull out the event history of a given entry based on either
        it's primary key value in the data store or its identifier field
        which is typically the textual identifier such as an EPC urn.

        .. code-block:: text

            # default format is json so format query parameter is optional
            http[s]:/[hostname]:[port]/epcis/events-by-entry-id/?format=[xml,json]

            # an example get reques for xml format
            http://localhost:8000/epcis/events-by-entry-id/urn:epc:id:sgtin:305555.05555sdf55.1/?format=xml

        :param request: The HTTP request
        :param format: json or xml
        :param entry_pk: The entry primary key.
        :param entry_identifier: The entry identifier.
        :return: A series of XML or JSON structures representing the events
        associated with the inbound id.

        '''
        args = {'epc_pk': entry_pk} if entry_pk else {
            'epc': entry_identifier}
        # get a list of EPCPyYes events from the DB proxy
        events = proxy.get_events_by_epc(**args)
        if len(events) > 0:
            epcis_document = template_events.EPCISEventListDocument(events)
            response_data = self.get_formatted_data(request, epcis_document,
                                                    format)
            return Response(response_data, status.HTTP_200_OK)
        else:
            raise NotFound(_('The entry with id %s could not be found.' % \
                             str(args)))


class EventsByILMDView(views.APIView, FormatHelperMixin):
    '''
    Gets all events associated with an ILMD name and value pair.
    For example Lot:2233
    '''

    def get(self, request: Request, format=None, ilmd_name=None,
            ilmd_value=None):
        pyyes_events = proxy.get_events_by_ilmd(ilmd_name, ilmd_value)
        if len(pyyes_events) > 0:
            epcis_document = template_events.EPCISEventListDocument(
                pyyes_events)
            response_data = self.get_formatted_data(request,
                                                    epcis_document,
                                                    format=format)
            return Response(response_data, status.HTTP_200_OK)
        else:
            msg = _('No events could be found that match name %s ' \
                    'and value %s' % (ilmd_name, ilmd_value))
            logger.debug(msg)
            raise NotFound(msg)


class MessageDetail(views.APIView, FormatHelperMixin):
    '''
    Returns a list of the full EPCIS messages that were received.
    '''

    def get(self, request: Request, format=None, message_id=None):
        '''
        Returns a full EPCIS message based on the inbpund message id.
        '''
        # get the message
        try:
            db_message = headers.Message.objects.get(id=message_id)
            message = proxy.get_full_message(db_message)
            return Response(self.get_formatted_data(request, message))
        except headers.Message.DoesNotExist:
            found_message_id = 'The message with id %s could not ' \
                               'be found' % message_id
            logger.debug()
            raise NotFound(found_message_id)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
import io
from quartet_capture.rules import Step
from quartet_epcis.parsing.parser import QuartetParser
from django.core.files.base import File
from quartet_capture.models import Rule, Step

def create_rule(apps, schema_editor):
    '''
    Creates the default rule.  Used in data-migrations.
    :return: None
    '''
    rule = Rule.objects.create(
        name='EPCIS',
        description='Auto-Created EPCIS Parsing Rule.',
    )
    Step.objects.create(
        name='Parse XML',
        description='Parse EPCIS data and save to database.',
        step_class='quartet_epcis.parsing.steps.EPCISParsingStep',
        order=1,
        rule=rule
    )


class EPCISParsingStep(Step):
    '''
    Calls the EPCIS parser as a rules.Step that can be used in the
    quartet_capture rule engine.
    '''
    def declared_parameters(self):
        return {}

    def execute(self, data, rule_context: dict):
        try:
            if isinstance(data, File):
                parser = QuartetParser(data)
            else:
                parser = QuartetParser(io.BytesIO(data))
        except TypeError:
            parser = QuartetParser(io.BytesIO(data.encode()))
        parser.parse()



# -*- coding: utf-8 -*-
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
from django.conf.urls import url
from quartet_epcis.routers import router
from quartet_epcis import views

app_name = 'quartet_epcis'

urlpatterns = [
    url(r'^event-detail/?$',
        views.EventDetailView.as_view(), name='event-detail'),
    url(r'^event-detail/(?P<event_id>[0-9a-fA-F\-_]{1,50})/$',
        views.EventDetailView.as_view(), name='event-detail'),
    url(r'^events-by-entry-id/?$',
        views.EntryEventHistoryView.as_view(),
        name='events-by-entry-id'),
    url(r'^events-by-entry-id/(?P<entry_identifier>[0-9a-zA-Z\W]{1,100})/$',
        views.EntryEventHistoryView.as_view(),
        name='events-by-entry-id'),
    url(r'^events-by-entry-pk/?$',
        views.EntryEventHistoryView.as_view(),
        name='events-by-entry-pk'),
    url(r'^events-by-entry-pk/(?P<entry_pk>[0-9a-zA-Z\W]{1,100})/$',
        views.EntryEventHistoryView.as_view(),
        name='events-by-entry-pk'),
    url(r'^events-by-ilmd/?$',
        views.EventsByILMDView.as_view(),
        name='events-by-ilmd'),
    url(r'^events-by-ilmd/(?P<ilmd_name>[[\w\s\W]{1,150})/'
        r'(?P<ilmd_value>[[\w\s\W]'
        r'{1,255})/$',
        views.EventsByILMDView.as_view(),
        name='events-by-ilmd'),
    url(r'^message/?$',
        views.MessageDetail.as_view(),
        name='message'
        ),
    url(r'^message/(?P<message_id>[\w-]{1,50})/$',
        views.MessageDetail.as_view(),
        name='message'
        ),
]

urlpatterns += router.urls



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

from django.db import models
from django.utils.translation import gettext_lazy as _
from model_utils import models as utils
from haikunator import Haikunator

haiku = Haikunator()

def haikunate():
    '''
    Since the haikunator is a class method
    it could not be used directly as a default callable for
    a django field...hence this function.
    '''
    return haiku.haikunate(token_length=4, token_hex=True)

document_type_choices = (
    ('Events', 'Events'),
    ('MasterData', 'MasterData'),
    ('QueryControl-Request', 'QueryControl-Request'),
    ('QueryControl-Response', 'QueryControl-Response'),
    ('QueryCallback', 'QueryCallback'),
    ('Query', 'Query')
)

partner_choices = (
    ('Sender', 'Sender'),
    ('Receiver', 'Receiver')
)


class Message(models.Model):
    '''
    An umbrella model to unify all headers and events that arrived
    as part of a single message.  If the message had a unique id supplied in
    the header, that will be used (albeit redunantly) along with a UUID4
    for each message as the primary key.
    '''
    id = models.CharField(
        max_length=50,
        null=False,
        help_text=_('The unique name of the job- autogenerated.'),
        verbose_name=_('Name'),
        unique=True,
        primary_key=True,
        default=haikunate
    )
    utils.AutoCreatedField(
        help_text=_('The date and time this record was created.'),
        verbose_name=_('Created Date')
    )

    class Meta:
        app_label = 'quartet_epcis'


class SBDH(models.Model):
    '''
    The db model representing the sbdh as defined by GS1.
    '''
    header_version = models.CharField(
        max_length=10,
        null=False,
        help_text=_('Descriptor which contains version '
                    'information for the SBDH.  Default is 1.0'),
        verbose_name=_('Header Version'),
        default='1.0'
    )
    document_identification = models.ForeignKey(
        'quartet_epcis.DocumentIdentification',
        null=False,
        on_delete=models.CASCADE
    )
    message = models.ForeignKey(
        'quartet_epcis.Message',
        on_delete=models.CASCADE,
        help_text=_('The message this header was associated with.'),
        verbose_name=_('Message')
    )

    class Meta:
        verbose_name = _('SBDH')
        verbose_name_plural = _('SBDHs')
        app_label = 'quartet_epcis'


class Partner(models.Model):
    '''
    Trading partner data.
    '''
    partner_type = models.CharField(
        max_length=20,
        null=False,
        help_text=_('The type of partner.  Either Sender or Receiver.'),
        verbose_name=_('Partner Type'),
        choices=partner_choices
    )
    authority = models.CharField(
        max_length=20,
        null=True,
        help_text=_('The identifying authority/format for the '
                    'identification field. Default is SGLN.'),
        verbose_name=_('authority'),
        default='SGLN'
    )
    identifier = models.CharField(
        max_length=100,
        null=True,
        help_text=_('An identifier that is in line with the authority '
                    'specified in the authority field.  Typically an '
                    'SGLN URN value.'),
        verbose_name=_('Identifier')
    )
    contact = models.CharField(
        max_length=50,
        null=True,
        help_text=_('The contact/name info.'),
        verbose_name=_('Contact')

    )
    email_address = models.EmailField(
        max_length=100,
        null=True,
        help_text=_('The email address for the partner.'),
        verbose_name=_('Email Address')
    )

    fax_number = models.CharField(
        max_length=20,
        null=True,
        help_text=_('Fax number.'),
        verbose_name=_('Fax Number')
    )
    telephone_number = models.CharField(
        max_length=20,
        null=True,
        help_text=_('Telephone number for the partner.'),
        verbose_name=_('Telephone number.')
    )
    contact_type_identifier = models.CharField(
        max_length=40,
        null=True,
        help_text=_('The type of contact- for example, '),
        verbose_name=_('Role of the contact in this business process.')
    )
    header = models.ForeignKey(
        'quartet_epcis.SBDH',
        on_delete=models.CASCADE,
        help_text=_('The related SBDH.'),
        verbose_name=_('SBDH')
    )

    def __str__(self):
        if self.authority:
            ret = '{0}:{1}'.format(self.authority, self.identifier)
        else:
            ret = '{0}:{1}'.format(self.id, self.partner_type)
        return ret

    class Meta:
        verbose_name = _('Partner')
        verbose_name_plural = _('Partners')
        app_label = 'quartet_epcis'


class DocumentIdentification(models.Model):
    '''
    The core model for the GS1 SBDH document identification element.
    '''
    standard = models.CharField(
        max_length=20,
        null=False,
        default='EPCglobal',
        help_text=_('The originator of the standard that the '
                    'following document falls under.  '
                    'Default is EPCglobal.'),
        verbose_name=_('Standard')
    )
    type_version = models.CharField(
        max_length=10,
        null=False,
        help_text=_('Descriptor which contains versioning information '
                    'or number of the standard that defines the document'
                    'which is specified in the ’Type’ data element, e.g. '
                    'values could be ‘1.3’ or ‘D.96A’, etc.'),
        verbose_name=_('Version'),
        default='1.0'
    )
    instance_identifier = models.CharField(
        max_length=100,
        null=False,
        help_text=_('Descriptor which contains reference information '
                    'which uniquely identifies this instance of the SBD '
                    'between the sender and the receiver.'),
        verbose_name=_('Instance Identifier'),
        db_index=True
    )
    document_type = models.CharField(
        max_length=25,
        null=False,
        help_text=_('A logical indicator representing the type of '
                    'Business Data being sent or the named type of '
                    'business data.'),
        verbose_name=_('Type'),
        choices=document_type_choices
    )
    multiple_type = models.NullBooleanField(
        default=False,
        help_text=_('A flag to indicate that there is more than one '
                    'type of Document in the instance.'),
        verbose_name=_('Multiple Type')
    )
    creation_date_and_time = models.CharField(
        max_length=35,
        null=True,
        help_text=_('Descriptor which contains date and time of '
                    'SBDH/document creation.'),
        verbose_name=_('Creation Date and Time')
    )

    def __str__(self):
        return 'Document Id {0}: {1}'.format(self.id,
                                             self.creation_date_and_time)

    class Meta:
        verbose_name = _('Document Identification')
        verbose_name_plural = _('Document Identifications')
        app_label = 'quartet_epcis'



# Generated by Django 2.0.3 on 2018-05-02 16:07

from django.db import migrations
from quartet_epcis.parsing.steps import create_rule


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.RunPython(create_rule)
    ]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

import uuid
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from quartet_epcis.models import choices
from haikunator import Haikunator
haiku = Haikunator()

def haikunate():
    '''
    Since the haikunator is a class method
    it could not be used directly as a default callable for
    a django field...hence this function.
    '''
    return haiku.haikunate(token_length=7, token_hex=True)

class UUIDModel(models.Model):
    '''
    A base model which uses UUIDs for primary key values.
    '''
    id = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
        help_text=_('Unique ID'),
        verbose_name=_('Unique ID'))

    class Meta:
        abstract = True


class SourceModel(models.Model):
    '''
    Abstract model class for any classes that reference a UUIDModel
    super-class.  Includes the UUID model UUID and a reference to the
    object type.
    '''
    source_event_id = models.UUIDField(
        help_text=_('The UUID of the source event within the '
                    'Quartet database.'),
        verbose_name=_('Source Event UUID'),
        null=False
    )
    source_event_type = models.CharField(
        max_length=2,
        choices=choices.EVENT_TYPE_CHOICES,
        help_text=_('The type of originating event: Object, Aggregation, '
                    'Transaction or Transformation.'),
        verbose_name=_('Source Event Type')
    )
    class Meta:
        abstract = True


class EPCISEvent(UUIDModel):
    '''
    The base EPCIS event as defined by GS1 on page 38 of the EPCIS 1.2 draft.
    '''
    event_time = models.DateTimeField(
        db_index=True,
        null=False,
        editable=False,
        help_text=_('The date and time at which the EPCIS Capturing '
                    'Application asserts the event occurred.'),
        verbose_name=_('Event Time')
    )
    event_timezone_offset = models.CharField(
        max_length=6,
        null=True,
        help_text=_('The time zone offset in effect at the '
                    'time and place the event occurred, expressed as an '
                    'offset from UTC'),
        verbose_name=_('Event Timezone Offset'),
        default='+00:00'
    )
    record_time = models.DateTimeField(
        null=True,
        help_text=_('The date and time at which this event was'
                    ' recorded by an EPCIS Repository.'),
        verbose_name=_('Record Time'),
        default=timezone.now
    )
    event_id = models.CharField(
        max_length=150,
        null=True,
        default=haikunate,
        help_text=_('An identifier for this event as specified by the '
                    'capturing application, globally unique across all events '
                    'other than error declarations. Not to be confused with '
                    'the unique id/primary key for events within a database.'),
        verbose_name=_('Event ID'),
        db_index=True
    )

    class Meta:
        abstract = True


class EPCISBusinessEvent(EPCISEvent):
    '''
    Abstract base-class for super-classes with an Action,
    biz step, biz location, etc...basically
    every main EPCIS class except the TransformationEvent class.
    '''
    action = models.CharField(
        max_length=5,
        choices=choices.ACTION_CHOICES,
        null=False,
        help_text=_('How this event relates to the lifecycle of the '
                    'EPCs named in this event.'),
        verbose_name=_('Action')
    )
    biz_step = models.CharField(
        max_length=150,
        null=True,
        help_text=_('The business step of which this event was a part.'),
        verbose_name=_('Business Step')
    )
    disposition = models.CharField(
        max_length=150,
        null=True,
        help_text=_('The business condition of the objects associated '
                    'with the EPCs, presumed to hold true until '
                    'contradicted by a subsequent event..'),
        verbose_name=_('Disposition')
    )
    read_point = models.CharField(
        max_length=150,
        null=True,
        help_text=_('The read point at which the event took place.'),
        verbose_name=_('Read Point')
    )
    biz_location = models.CharField(
        max_length=150,
        null=True,
        help_text=_('The business location where the objects '
                    'associated with the EPCs may be found, '
                    'until contradicted '
                    'by a subsequent event.'),
        verbose_name=_('Business Location')
    )

    class Meta:
        abstract = True



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

from rest_framework import viewsets
from quartet_epcis.models import events, entries, headers
from quartet_epcis import serializers

class EntryViewSet(viewsets.ModelViewSet):
    '''
    The default viewset to handle the management of Entries.
    '''
    queryset = entries.Entry.objects.all()
    serializer_class = serializers.EntrySerializer
    search_fields = ['identifier']

class EntryEventViewSet(viewsets.ModelViewSet):
    '''
    The default viewset to handle EntryEvents.
    '''
    queryset = entries.EntryEvent.objects.all()
    serializer_class = serializers.EntrySerializer


class EventViewSet(viewsets.ModelViewSet):
    '''
    The default viewset to handle Events.
    '''
    queryset = events.Event.objects.all()
    serializer_class = serializers.EventSerializer
    search_fields = ['id', 'biz_step', 'biz_location', 'event_id',
                     'event_time', 'record_time', 'read_point',
                     'action', 'disposition', 'message_id']
    filter_fields = '__all__'


class TransformationIDViewSet(viewsets.ModelViewSet):
    '''
    The default viewset to handle TransformationIDs.
    '''
    queryset = events.TransformationID.objects.all()
    serializer_class = serializers.TransformationIDSerializer


class ErrorDeclarationViewSet(viewsets.ModelViewSet):
    '''
    The default viewset to handle ErrorDeclarations.
    '''
    queryset = events.ErrorDeclaration.objects.all()
    serializer_class = serializers.ErrorDeclarationSerializer


class QuantityElementViewSet(viewsets.ModelViewSet):
    '''
    The default viewset to handle QuantityElements.
    '''
    queryset = events.QuantityElement.objects.all()
    serializer_class = serializers.QuantityElementSerializer


class BusinessTransactionViewSet(viewsets.ModelViewSet):
    '''
    The default viewset to handle BusinessTransactions.
    '''
    queryset = events.BusinessTransaction.objects.all()
    serializer_class = serializers.BusinessTransactionSerializer


class InstanceLotMasterDataViewSet(viewsets.ModelViewSet):
    '''
    The default viewset to handle InstanceLotMasterData.
    '''
    queryset = events.InstanceLotMasterData.objects.all()
    serializer_class = serializers.InstanceLotMasterDataSerializer


class SourceViewSet(viewsets.ModelViewSet):
    '''
    The default viewset to handle Sources.
    '''
    queryset = events.Source.objects.all()
    serializer_class = serializers.SourceSerializer


class SourceEventViewSet(viewsets.ModelViewSet):
    '''
    The default viewset to handle SourceEvents.
    '''
    queryset = events.SourceEvent.objects.all()
    serializer_class = serializers.SourceEventSerializer


class DestinationViewSet(viewsets.ModelViewSet):
    '''
    The default viewset to handle Destinations.
    '''
    queryset = events.Destination.objects.all()
    serializer_class = serializers.DestinationSerializer


class DestinationEventViewSet(viewsets.ModelViewSet):
    '''
    The default viewset to handle DestinationEvents.
    '''
    queryset = events.DestinationEvent.objects.all()
    serializer_class = serializers.DestinationEventSerializer


class MessageViewSet(viewsets.ReadOnlyModelViewSet):
    '''
    Default view for messages.
    '''
    queryset = headers.Message.objects.all()
    serializer_class = serializers.MessageSerializer
